// Copyright (c) 2010-2023, Lawrence Livermore National Security, LLC. Produced
// at the Lawrence Livermore National Laboratory. All Rights reserved. See files
// LICENSE and NOTICE for details. LLNL-CODE-806117.
//
// This file is part of the MFEM library. For more information and source code
// availability visit https://mfem.org.
//
// MFEM is free software; you can redistribute it and/or modify it under the
// terms of the BSD-3 license. We welcome feedback and contributions, see file
// CONTRIBUTING.md for details.

#include "osockstream.hpp"
#include "globals.hpp"

namespace mfem
{

osockstream::osockstream(int port, const char *hostname)
   : socketstream(hostname, port)
{
   if (!is_open())
   {
      mfem::err << "Unable to connect to port " << port << " on "
                << hostname << '\n';
   }
}

}
